from django.contrib import admin
from reversion.admin import VersionAdmin

from huscy.projects import models
from huscy.projects.services import create_project


@admin.register(models.ResearchUnit)
class ResearchUnitAdmin(VersionAdmin, admin.ModelAdmin):
    list_display = 'name', 'code', 'principal_investigator'
    search_fields = (
        'name',
        'principal_investigator__first_name',
        'principal_investigator__last_name',
    )


@admin.register(models.Project)
class ProjectAdmin(VersionAdmin, admin.ModelAdmin):
    list_display = 'id', 'local_id_name', 'title', 'principal_investigator'
    search_fields = (
        'principal_investigator__first_name',
        'principal_investigator__last_name',
        'title',
    )

    def save_model(self, request, project, form, change):
        if change:
            super().save_model(request, project, form, change)
        else:
            create_project(project.title, project.research_unit, project.principal_investigator,
                           creator=request.user, local_id=project.local_id,
                           description=project.description)


@admin.register(models.Membership)
class MembershipAdmin(VersionAdmin, admin.ModelAdmin):
    list_display = 'id', '_project', 'user', 'is_coordinator', 'has_write_permission'
    list_filter = 'is_coordinator',
    search_fields = 'project__title', 'user__first_name', 'user__last_name'

    def has_add_permission(self, request, obj=None):
        return False

    def has_change_permission(self, request, obj=None):
        return False

    def _project(self, membership):
        project = membership.project
        return f'{project.id} ({project.local_id_name} {project.title})'



from uuid import uuid4

from django.conf import settings
from django.db import models
from django.utils.translation import gettext_lazy as _


def get_default_for_research_unit_code():
    return uuid4().hex


class ResearchUnit(models.Model):
    name = models.CharField(_('Name'), max_length=128, unique=True)
    code = models.CharField(_('Code'), max_length=32, unique=True, blank=True,
                            default=get_default_for_research_unit_code)
    principal_investigator = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.PROTECT,
                                               verbose_name=_('Principal investigator'))

    local_id_start_value = models.PositiveIntegerField(_('Local ID start value'),
                                                       default=1)

    def __str__(self):
        return f'{self.name} ({self.principal_investigator.get_full_name()})'

    class Meta:
        ordering = 'name',
        verbose_name = _('Research unit')
        verbose_name_plural = _('Research units')


class Project(models.Model):
    research_unit = models.ForeignKey(ResearchUnit, on_delete=models.PROTECT,
                                      verbose_name=_('Research unit'))
    local_id = models.PositiveIntegerField(_('Local ID'))

    title = models.CharField(_('Title'), max_length=255)
    description = models.TextField(_('Description'), blank=True, default='')

    principal_investigator = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.PROTECT,
                                               related_name='+',
                                               verbose_name=_('Principal investigator'))

    @property
    def local_id_name(self):
        return f'{self.research_unit.code}-{self.local_id}'

    class Meta:
        ordering = '-id',
        unique_together = ('local_id', 'research_unit')
        verbose_name = _('Project')
        verbose_name_plural = _('Projects')


class Membership(models.Model):
    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE,
                             verbose_name=_('User'))
    project = models.ForeignKey(Project, on_delete=models.CASCADE, verbose_name=_('Project'))
    is_coordinator = models.BooleanField(_('Is coordinator'), default=False)

    @property
    def has_write_permission(self):
        return self.user.has_perm('change_project', self.project)

    def __str__(self):
        return f'{self.user.get_full_name()} is member in project {self.project.title}'

    class Meta:
        ordering = 'project', 'user'
        unique_together = 'user', 'project'
        verbose_name = _('Membership')
        verbose_name_plural = _('Memberships')



from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('projects', '0001_initial'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='membership',
            options={'ordering': ('project', 'user'), 'verbose_name': 'Membership', 'verbose_name_plural': 'Memberships'},
        ),
    ]



from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import huscy.projects.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='ResearchUnit',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=128, unique=True, verbose_name='Name')),
                ('code', models.CharField(blank=True, default=huscy.projects.models.get_default_for_research_unit_code, max_length=32, unique=True, verbose_name='Code')),
                ('principal_investigator', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to=settings.AUTH_USER_MODEL, verbose_name='Principal investigator')),
            ],
            options={
                'verbose_name': 'Research unit',
                'verbose_name_plural': 'Research units',
                'ordering': ('name',),
            },
        ),
        migrations.CreateModel(
            name='Project',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('local_id', models.PositiveIntegerField(verbose_name='Local ID')),
                ('title', models.CharField(max_length=255, verbose_name='Title')),
                ('description', models.TextField(blank=True, default='', verbose_name='Description')),
                ('principal_investigator', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to=settings.AUTH_USER_MODEL, verbose_name='Principal investigator')),
                ('research_unit', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='projects.researchunit', verbose_name='Research unit')),
            ],
            options={
                'verbose_name': 'Project',
                'verbose_name_plural': 'Projects',
                'ordering': ('-id',),
                'unique_together': {('local_id', 'research_unit')},
            },
        ),
        migrations.CreateModel(
            name='Membership',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('is_coordinator', models.BooleanField(default=False, verbose_name='Is coordinator')),
                ('project', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='projects.project', verbose_name='Project')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL, verbose_name='User')),
            ],
            options={
                'verbose_name': 'Membership',
                'verbose_name_plural': 'Memberships',
                'unique_together': {('user', 'project')},
            },
        ),
    ]



from itertools import count, filterfalse

from django.db.models import Q
from django.db.transaction import atomic
from guardian.shortcuts import assign_perm, remove_perm

from huscy.projects.models import Membership, Project, ResearchUnit


@atomic
def create_membership(project, user, is_coordinator=False, has_write_permission=False):
    assign_perm('view_project', user, project)
    if is_coordinator or has_write_permission:
        assign_perm('change_project', user, project)
    return Membership.objects.create(
        project=project,
        user=user,
        is_coordinator=is_coordinator,
    )


@atomic
def create_project(title, research_unit, principal_investigator, creator,
                   local_id=None, description=''):
    if local_id is None:
        local_id = _get_next_local_id(research_unit)

    project = Project.objects.create(
        description=description,
        local_id=local_id,
        principal_investigator=principal_investigator,
        research_unit=research_unit,
        title=title,
    )

    create_membership(project, principal_investigator, is_coordinator=True)
    if principal_investigator != creator:
        create_membership(project, creator, is_coordinator=True)

    return project


def _get_next_local_id(research_unit):
    taken_values = (Project.objects.filter(research_unit=research_unit)
                                   .values_list('local_id', flat=True))
    return next(filterfalse(lambda x: x in set(taken_values),
                            count(research_unit.local_id_start_value)))


@atomic
def delete_membership(membership):
    remove_perm('view_project', membership.user, membership.project)
    remove_perm('change_project', membership.user, membership.project)
    membership.delete()


@atomic
def delete_project(project):
    map(delete_membership, project.membership_set.all())
    project.delete()


def get_memberships(project):
    return Membership.objects.filter(project=project)


def get_participating_projects(user):
    return (Project.objects
                   .filter(Q(principal_investigator=user) | Q(membership__user=user))
                   .distinct())


def get_projects():
    return Project.objects.all()


def get_research_units():
    return ResearchUnit.objects.all()


def is_project_coordinator(project, user):
    return Membership.objects.filter(project=project, user=user, is_coordinator=True).exists()


def is_project_member_with_write_permission(project, user):
    return all([
        Membership.objects.filter(project=project, user=user).exists(),
        user.has_perm('projects.change_project', project),
    ])


def set_principal_investigator(project, user):
    if not Membership.objects.filter(project=project, user=user, is_coordinator=True).exists():
        raise ValueError('Only project members who are coordinators can become '
                         'principal investigators.')

    project.principal_investigator = user
    project.save(update_fields=['principal_investigator'])
    return project


def update_membership(membership, is_coordinator, has_write_permission):
    if is_coordinator or has_write_permission:
        assign_perm('change_project', membership.user, membership.project)
    else:
        remove_perm('change_project', membership.user, membership.project)
    membership.is_coordinator = is_coordinator
    membership.save()
    return membership


def update_project(project, local_id=None, title=None, description=None):
    update_fields = []
    if local_id is not None:
        project.local_id = local_id
        update_fields.append('local_id')
    if title is not None:
        project.title = title
        update_fields.append('title')
    if description is not None:
        project.description = description
        update_fields.append('description')
    project.save(update_fields=update_fields)
    return project



from rest_framework.permissions import BasePermission, SAFE_METHODS
from huscy.projects.services import is_project_coordinator, is_project_member_with_write_permission


class ReadOnly(BasePermission):
    def has_permission(self, request, view):
        return request.user.is_authenticated and request.method in SAFE_METHODS


class ChangeProjectPermission(BasePermission):
    def has_object_permission(self, request, view, project):
        if request.method in ['PATCH', 'PUT']:
            return any([
                request.user.has_perm('projects.change_project'),
                is_project_member_with_write_permission(project, request.user),
            ])
        return True


class CreateProjectPermission(BasePermission):
    pass


class DeleteProjectPermission(BasePermission):
    def has_object_permission(self, request, view, project):
        if request.method == 'DELETE':
            return any([
                request.user.has_perm('projects.delete_project'),
                request.user.has_perm('projects.delete_project', project),
            ])
        return True


class IsProjectCoordinator(BasePermission):
    def has_permission(self, request, view):
        if request.user.is_superuser:
            return True
        return is_project_coordinator(view.project, request.user)

    def has_object_permission(self, request, view, instance):
        if request.user.is_superuser:
            return True
        return is_project_coordinator(view.project, request.user)



import logging

from django.db.models import Q
from django.shortcuts import get_object_or_404
from rest_framework import filters, mixins, viewsets
from rest_framework.decorators import action
from rest_framework.permissions import DjangoModelPermissions, IsAuthenticated
from rest_framework.response import Response
from reversion import set_comment
from reversion.views import RevisionMixin

from huscy.projects import serializers, services
from huscy.projects.models import Project
from huscy.projects.permissions import (
    ChangeProjectPermission,
    CreateProjectPermission,
    DeleteProjectPermission,
    IsProjectCoordinator,
    ReadOnly,
)

logger = logging.getLogger('projects')


class MembershipViewSet(RevisionMixin, mixins.CreateModelMixin, mixins.DestroyModelMixin,
                        mixins.ListModelMixin, mixins.UpdateModelMixin, viewsets.GenericViewSet):
    serializer_class = serializers.MembershipSerializer
    permission_classes = (IsAuthenticated, IsProjectCoordinator | ReadOnly)

    def initial(self, request, *args, **kwargs):
        self.project = get_object_or_404(Project, pk=self.kwargs['project_pk'])
        super().initial(request, *args, **kwargs)

    def get_queryset(self):
        return services.get_memberships(self.project)

    def perform_create(self, serializer):
        membership = serializer.save(project=self.project)
        set_comment(f'Created membership <ID-{membership.id}> for project <ID-{self.project.id}>')

        logger.info('Membership created by user %s for project "%s" and member %s',
                    self.request.user.username, membership.project.title,
                    membership.user.get_full_name())

    def perform_destroy(self, membership):
        services.delete_membership(membership)
        set_comment(f'Deleted membership <ID-{membership.id}>')

        logger.info('Membership deleted by user %s for project "%s" and member %s',
                    self.request.user.username, membership.project.title,
                    membership.user.get_full_name())

    def perform_update(self, serializer):
        membership = serializer.save()
        set_comment(f'Updated membership <ID-{membership.id}>')

        logger.info('Membership updated by user %s for project "%s" and member %s',
                    self.request.user.username, membership.project.title,
                    membership.user.get_full_name())


class ViewProjectFilter(filters.BaseFilterBackend):
    def filter_queryset(self, request, queryset, view):
        user = request.user
        if user.has_perm('projects.view_project'):
            return queryset
        return queryset.filter(Q(principal_investigator=user) | Q(membership__user=user))


class ProjectViewSet(RevisionMixin, viewsets.ModelViewSet):
    filter_backends = ViewProjectFilter,
    permission_classes = (IsAuthenticated, ChangeProjectPermission, CreateProjectPermission,
                          DeleteProjectPermission)
    queryset = services.get_projects()

    def initial(self, request, *args, **kwargs):
        if self.action == 'principalinvestigator':
            self.project = get_object_or_404(Project, pk=self.kwargs['pk'])
        super().initial(request, *args, **kwargs)

    def get_serializer_class(self):
        if self.action == 'create':
            return serializers.CreateProjectSerializer
        return serializers.ProjectSerializer

    def perform_create(self, serializer):
        project = serializer.save()
        set_comment(f'Created project <ID-{project.id}>')

        logger.info('Project created by user %s with title "%s"',
                    self.request.user.username, project.title)

    def perform_destroy(self, project):
        services.delete_project(project)
        set_comment(f'Deleted project <ID-{project.id}>')

        logger.info('Project deleted by user %s with title "%s"',
                    self.request.user.username, project.title)

    def perform_update(self, serializer):
        project = serializer.save()
        set_comment(f'Updated project <ID-{project.id}>')

        logger.info('Project updated by user %s with title "%s"',
                    self.request.user.username, project.title)

    @action(detail=True, methods=['PUT'],
            permission_classes=[IsAuthenticated, IsProjectCoordinator])
    def principalinvestigator(self, request, pk):
        serializer = serializers.PrincipalInvestigatorSerializer(instance=self.get_object(),
                                                                 data=request.data)
        serializer.is_valid(raise_exception=True)
        project = serializer.save()

        set_comment(f'Changed principal investigator for project <ID-{project.id}>')
        return Response(serializers.ProjectSerializer(project, context=dict(request=request)).data)


class ResearchUnitViewSet(RevisionMixin, viewsets.ModelViewSet):
    permission_classes = (DjangoModelPermissions | ReadOnly, )
    queryset = services.get_research_units()
    serializer_class = serializers.ResearchUnitSerializer

    def perform_create(self, serializer):
        research_unit = serializer.save()
        set_comment(f'Created research unit <ID-{research_unit.id}>')

    def perform_destroy(self, research_unit):
        research_unit.delete()
        set_comment(f'Deleted research unit <ID-{research_unit.id}>')

    def perform_update(self, serializer):
        research_unit = serializer.save()
        set_comment(f'Updated research unit <ID-{research_unit.id}>')



from django.db.utils import IntegrityError
from rest_framework import serializers
from rest_framework.exceptions import ValidationError

from huscy.projects.models import Membership, Project, ResearchUnit
from huscy.projects.services import (
    create_membership,
    create_project,
    set_principal_investigator,
    update_membership,
    update_project,
)


class ResearchUnitSerializer(serializers.ModelSerializer):
    class Meta:
        model = ResearchUnit
        fields = (
            'id',
            'code',
            'name',
            'principal_investigator',
        )


class PrincipalInvestigatorSerializer(serializers.ModelSerializer):
    class Meta:
        model = Project
        fields = 'principal_investigator',

    def update(self, project, validated_data):
        principal_investigator = validated_data.get('principal_investigator')
        try:
            return set_principal_investigator(project=project, user=principal_investigator)
        except ValueError as e:
            raise ValidationError({'detail': e})


class ProjectSerializer(serializers.ModelSerializer):

    class Meta:
        model = Project
        fields = (
            'id',
            'description',
            'local_id',
            'local_id_name',
            'participating',
            'principal_investigator',
            'principal_investigator_name',
            'research_unit',
            'research_unit_name',
            'title',
        )
        extra_args = {
            'description': {'required': False},
            'local_id': {'required': False},
            'title': {'required': False},
        }
        read_only_fields = 'principal_investigator', 'research_unit'

    participating = serializers.SerializerMethodField()
    principal_investigator_name = serializers.CharField(
        source='principal_investigator.get_full_name', read_only=True,
    )
    research_unit_name = serializers.CharField(source='research_unit.name', read_only=True)

    # default=None is used as required=False does not work together with the unique_together
    # constraint, however this defaut=None is a workaround, see:
    # https://github.com/encode/django-rest-framework/issues/4456
    local_id = serializers.IntegerField(min_value=1, default=None)

    def get_participating(self, project):
        user = self.context['request'].user
        return (project.principal_investigator == user or
                project.membership_set.filter(user=user).exists())

    def update(self, project, validated_data):
        try:
            return update_project(project=project, **validated_data)
        except IntegrityError:
            raise ValidationError({'local_id': ['Local ID already exists.']})


class CreateProjectSerializer(ProjectSerializer):
    creator = serializers.HiddenField(default=serializers.CurrentUserDefault())

    class Meta:
        model = Project
        fields = (
            'id',
            'creator',
            'description',
            'local_id',
            'local_id_name',
            'participating',
            'principal_investigator',
            'principal_investigator_name',
            'research_unit',
            'research_unit_name',
            'title',
        )

    def create(self, validated_data):
        return create_project(**validated_data)


class MembershipSerializer(serializers.ModelSerializer):
    has_write_permission = serializers.BooleanField()
    username = serializers.CharField(source='user.get_full_name', read_only=True)

    class Meta:
        model = Membership
        fields = (
            'id',
            'has_write_permission',
            'is_coordinator',
            'project',
            'user',
            'username',
        )
        read_only_fields = 'project',

    def create(self, validated_data):
        return create_membership(**validated_data)

    def update(self, membership, validated_data):
        return update_membership(membership, validated_data.get('is_coordinator'),
                                 validated_data.get('has_write_permission'))



from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('projects', '0004_remove_project_project_manager'),
    ]

    operations = [
        migrations.AddField(
            model_name='researchunit',
            name='local_id_start_value',
            field=models.PositiveIntegerField(default=1, verbose_name='Local ID start value'),
        ),
    ]



from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('projects', '0002_alter_membership_options'),
    ]

    operations = [
        migrations.AddField(
            model_name='project',
            name='project_manager',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.PROTECT, related_name='+', to=settings.AUTH_USER_MODEL, verbose_name='Project manager'),
        ),
        migrations.AlterField(
            model_name='project',
            name='principal_investigator',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='+', to=settings.AUTH_USER_MODEL, verbose_name='Principal investigator'),
        ),
    ]



# Generated by Django 5.0.7 on 2025-01-31 03:33

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("projects", "0003_project_project_manager_and_more"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="project",
            name="project_manager",
        ),
    ]
